
/* kgpio.c -- look at GPIO lines.
	Needed along with SPI (SPIDEV) and i2c
	in combinations */


/* Uses the /dev/mem addresses to change GPIO lines. 
	Based to some extent on code seen in 

	http://elinux.org/RPi_Low-level_peripherals

	But there really ought to be some real device driver... 

	Further notes based on the datasheet published for the BCM2835.
*/


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdint.h>
#include "kgpio.h"


/* Available gpios on the P1 and P5 headers: */ 
struct tag_gpioinfo thegpinfo[] = 
{    /*                                                                                                                   9876543210 */
{ 0, "NA-00", {"input", "output", "F5=.DU", "F4=.DU", "F0=I2C0_SDA",  "F1=SA5",   "F2=.RE",  "F3=.DU" }, 	 0,          00000000007, 0, 0x00000001 },  /* 0 */
{ 1, "NA-00", {"input", "output", "F5=.DU", "F4=.DU", "F0=I2C0_SDA",  "F1=SA4",   "F2=.RE",  "F3=.DU" }, 	 0,          00000000070, 0, 0x00000002 },  /* 1 */
{ 2, "P1-03", {"input", "output", "F5=.DU", "F4=.DU", "F0=I2C1_SDA",  "F1=SA3",   "F2=.RE",  "F3=.DU" }, 	 0,          00000000700, 0, 0x00000004 },  /* 2 */
{ 3, "P1-05", {"input", "output", "F5=.DU", "F4=.DU", "F0=I2C1_SCL",  "F1=SA2",   "F2=.RE",  "F3=.DU" }, 	 0,          00000007000, 0, 0x00000008 },  /* 3 */
{ 4, "P1-07", {"input", "output", "F5=ARM_TDI","F4=.DU","F0=GPCLK0",   "F1=SA1",   "F2=.RE",  "F3=.DU" }, 	 0,          00000070000, 0, 0x00000010 },  /* 4 */
{ 5, "NA-00", {"input", "output", "F5=ARM_TDO","F4=.DU","F0=GPCLK1",   "F1=SA0",   "F2=.RE",  "F3=.DU" }, 	 0,          00000700000, 0, 0x00000020 },  /* 5 */
{ 6, "NA-00", {"input", "output", "F5=ARM_RTCK","F4=.DU","F0=GPCLK2","F1=SOE_N/SE","F2=.RE",  "F3=.DU" }, 	 0,          00007000000, 0, 0x00000040 },  /* 6 */
{ 7, "P1-26", {"input", "output", "F5=.DU", "F4=.DU","F0=SPI0_CE1_N","F1=SWE_N/SRW_N","F2=.RE","F3=.DU"},    0,          00070000000, 0, 0x00000080 },  /* 7 */
{ 8, "P1-24", {"input", "output", "F5=.DU", "F4=.DU", "F0=SPI0_CE0_N","F1=SD0",   "F2=.RE",  "F3=.DU" },     0,          00700000000, 0, 0x00000100 },  /* 8 */
{ 9, "P1-21", {"input", "output", "F5=.DU", "F4=.DU", "F0=SPI0_MISO", "F1=SD1",   "F2=.RE",  "F3=.DU" },     0,          07000000000, 0, 0x00000200 },  /* 9 */
{10, "P1-19", {"input", "output", "F5=.DU", "F4=.DU", "F0=SPI0_MOSI", "F1=SD2",   "F2=.RE",  "F3=.DU" },     1,          00000000007, 0, 0x00000400 },  /* 10 */
{11, "P1-23", {"input", "output", "F5=.DU", "F4=.DU", "F0=SPI0_SCLK", "F1=SD3",   "F2=.RE",  "F3=.DU" },     1,          00000000070, 0, 0x00000800 },  /* 11 */
{12, "NA-00", {"input", "output", "F5=ARM_TMS","F4=.DU","F0=PWM0",     "F1=SD4",   "F2=.RE",  "F3=.DU" }, 	 1,          00000000700, 0, 0x00001000 },  /* 12 */
{13, "NA-00", {"input", "output", "F5=ARM_TCK","F4=.DU","F0=PWM1",     "F1=SD5",   "F2=.RE",  "F3=.DU" }, 	 1,          00000007000, 0, 0x00002000 },  /* 13 */
{14, "P1-08", {"input", "output", "F5=TXD1", "F4=.DU","F0=UART0_TXD",   "F1=SD6",   "F2=.RE", "F3=.DU" },    1,          00000070000, 0, 0x00004000 },  /* 14 */
{15, "P1-10", {"input", "output", "F5=RXD1", "F4=.DU","F0=UART0_RXD",   "F1=SD7",   "F2=.RE", "F3=.DU" },    1,          00000700000, 0, 0x00008000 },  /* 15 */
{16, "NA-00", {"input", "output", "F5=CTS1", "F4=SPI1_CE2_N","F0=.RE", "F1=SD8",    "F2=.RE", "F3=CTS0" },   1,          00007000000, 0, 0x00010000 },  /* 16 */
{17, "P1-11", {"input", "output", "F5=UART1_RTS","F4=SPI1_CE1_N","F0=.RE","F1=SD9",  "F2=.RE", "F3=RTS0" },  1,          00070000000, 0, 0x00020000 },  /* 17 */
{18, "P1-12", {"input", "output", "F5=PWM0", "F4=SPI1_CE0_N", "F0=PCM_CLK", "F1=SD10", "F2=.RE", "F3=BSCSL SDA/MOSI" },1,00700000000, 0, 0x00040000 },  /* 18 */
{19, "NA-00", {"input", "output", "F5=PWM1", "F4=SPI1_MISO",   "F0=PCM_FS", "F1=SD11", "F2=.RE","F3=BSCSL SCL/SCLK"},1,  07000000000, 0, 0x00080000 },  /* 19 */
{20, "NA-00", {"input", "output", "F5=GPCLK0","F4=SPI1_MOSI", "F0=PCM_DIN", "F1=SD12","F2=.RE","F3=BSCSL -/MISO" }, 2,   00000000007, 0, 0x00100000 },  /* 20 */
{21, "NA-00", {"input", "output", "F5=GPCLK1","F4=SPI1_SCLK", "F0=PCM_DOUT","F1=SD13","F2=.RE","F3=BSCSL -/CE_N" }, 2,   00000000070, 0, 0x00200000 },  /* 21 */
{22, "P1-15", {"input", "output", "F5=.DU",  "F4=ARM_TRST", "F0=.RE", "F1=SD14",  "F2=.RE",  "F3=SD1_CLK" },  2,         00000000700, 0, 0x00400000 },  /* 22 */
{23, "P1-16", {"input", "output", "F5=.DU",  "F4=ARM_RTCK", "F0=.RE", "F1=SD15",  "F2=.RE",  "F3=SD1_CMD" },  2,         00000007000, 0, 0x00800000 },  /* 23 */
{24, "P1-18", {"input", "output", "F5=.DU",  "F4=ARM_TDO",  "F0=.RE", "F1=SD16",  "F2=.RE",  "F3=SD1_DATA0" },2,         00000070000, 0, 0x01000000 },  /* 24 */
{25, "P1-22", {"input", "output", "F5=.DU",  "F4=ARM_TCK",  "F0=.RE", "F1=SD17",  "F2=.RE",  "F3=SD1_DAT1" }, 2,         00000700000, 0, 0x02000000 },  /* 25 */
{26, "NA-00", {"input", "output", "F5=.DU",  "F4=ARM_TDI",  "F0=.RE", "F1=.RE",   "F2=.RE",  "F3=SD1_DAT2" }, 2,         00007000000, 0, 0x04000000 },  /* 26 */
{27, "P1-13", {"input", "output", "F5=.DU",  "F4=ARM_TMS",  "F0=.RE", "F1=.RE",   "F2=.RE",  "F3=SDI_DAT3" }, 2,         00070000000, 0, 0x08000000 },  /* 27 */
{28, "P5-03", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SDA0",    "F1=SA5",   "F2=PCM_CLK", "F3=.RE" },   2,         00700000000, 0, 0x10000000 },  /* 28 */
{29, "P5-04", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SCL0",    "F1=SA4",   "F2=PCM_FS",  "F3=.RE" },   2,         07000000000, 0, 0x20000000 },  /* 29 */
{30, "P5-05", {"input", "output", "F5=CTS1", "F4=.DU",  "F0=.RE",     "F1=SA3",   "F2=PCM_DIN", "F3=CTS0" },  3,         00000000007, 0, 0x40000000 },  /* 30 */
{31, "P5-06", {"input", "output", "F5=RTS1", "F4=.DU",  "F0=.RE",     "F1=SA2",   "F2=PCM_DOUT", "F3=RTS0" }, 3,         00000000070, 0, 0x80000000 },  /* 31 */
{32, "NA-00", {"input", "output", "F5=TXD1",  "F4=.DU",  "F0=GPCLK0", "F1=SA1",   "F2=.RE",  "F3=TXD0" },     3,         00000000700, 1, 0x00000001 },  /* 32 */
{33, "NA-00", {"input", "output", "F5=RXD1",  "F4=.DU",  "F0=.RE",    "F1=SA0",   "F2=.RE",  "F3=RXD0" },     3,         00000007000, 1, 0x00000002 },  /* 33 */
{34, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=GPCLK0",    "F1=SOE_N/SE",   "F2=.RE",  "F3=.RE" }, 3,       00000070000, 1, 0x00000004 },  /* 34 */
{35, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SPI0_CE1_N", "F1=SWE_N/SRW_N","F2=.DU",  "F3=.RE" },3,      00000700000, 1, 0x00000008 },  /* 35 */
{36, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SPI0_CE0_N", "F1=SD0",   "F2=TXD0",  "F3=.RE" },  3,         00007000000, 1, 0x00000010 },  /* 36 */
{37, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SPI0_MISO",  "F1=SD1",   "F2=RXD0",  "F3=.RE" },  3,         00070000000, 1, 0x00000020 },  /* 37 */
{38, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SPI0_MOSI",  "F1=SD2",   "F2=RTS0",  "F3=.RE" },  3,         00700000000, 1, 0x00000040 },  /* 38 */
{39, "NA-00", {"input", "output", "F5=.DU",  "F4=.DU",  "F0=SPI0_SCLK",  "F1=SD3",   "F2=CTS0",  "F3=.RE" },  3,         07000000000, 1, 0x00000080 },  /* 39 */
{40, "NA-00", {"input", "output", "F5=TXD1","F4=SPI2_MISO","F0=PWM0",    "F1=SD4",   "F2=.DU",  "F3=.RE" },   4,         00000000007, 1, 0x00000100 },  /* 40 */

{41, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00000000070, 1, 0x00000200 },  /* 41 */
{42, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00000000700, 1, 0x00000400 },  /* 42 */
{43, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00000007000, 1, 0x00000800 },  /* 43 */
{44, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00000070000, 1, 0x00001000 },  /* 44 */
{45, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00000700000, 1, 0x00002000 },  /* 45 */
{46, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00007000000, 1, 0x00004000 },  /* 46 */
{47, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00070000000, 1, 0x00008000 },  /* 47 */
{48, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         00700000000, 1, 0x00010000 },  /* 48 */
{49, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        4,         07000000000, 1, 0x00020000 },  /* 49 */
{50, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        5,         00000000007, 1, 0x00040000 },  /* 50 */
{51, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        5,         00000000070, 1, 0x00080000 },  /* 51 */
{52, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        5,         00000000700, 1, 0x00100000 },  /* 52 */
{53, "NA-00", {"input", "output", "F5=.XX",  "F4=.DU",  "F0=.RE",    "F1=.XX",   "F2=.RE",  "F3=.XX" },        5,         00000007000, 1, 0x00200000 }   /* 53 */
}; /* And so on, we don't really worry too much about the ones past 32, until we have access to some of the wires. */

/* Useable on the P1 and P5 pin headers. */
int GPuseable[] = {2, 3, 4, 7, 8, 9, 10, 11, 14, 15, 17, 18, 22, 23, 24, 25, 27, 28, 29, 30, 31};

/* Find the base-pointer 
	Returns 0 for OK, -1 on failures.
*/

int GPbasepointer(volatile uint32_t **pgpio)
{
	int mem_fd;

	char *gpio_map;

	*pgpio = NULL;

	mem_fd = open("/dev/mem", O_RDWR | O_SYNC);
	if(mem_fd < 0)
	{
		return(mem_fd);
	}
	
	gpio_map = (char *) mmap(NULL, 
		GPBLOCK_SIZE, 
		PROT_READ|PROT_WRITE,
		MAP_SHARED,
		mem_fd,
		GPIO_BASE); 

	close(mem_fd);

	*pgpio = (volatile uint32_t *) gpio_map;

	return(0);
}



/* Read the capabilty word for the gpio-line identified as channel. */

int GPreadcap(volatile uint32_t *gpio, int channel)
{
	uint32_t rg0;
	unsigned c10, c1;

	if(channel < 0) return(-1);
	if(channel > 53) return(-1);

	c10 = channel / 10;
	c1 = channel % 10;

	rg0 = *(gpio + c10);
	rg0 >>= (3*c1);
	rg0 &= 7;
	return(rg0);
}


/* Set the function on the indicated channel to the value. Beware of
	illegal values ... most will only allow input or output. */

int GPsetcap(volatile uint32_t *gpio, int channel, int capability)
{
	uint32_t rg0, pat, om;
	unsigned c10, c1;

	if(channel < 0) return(-1);
	if(channel > 53) return(-1);

	c10 = channel / 10;
	c1 = channel % 10;

	om = 7 << (c1*3);
	pat = (capability & 7) << (c1*3);

	rg0 = *(gpio + c10);
	rg0 &= ~om;
	rg0 |= pat;
	*(gpio + c10) = rg0;

	return(rg0);
}


int GPshowone(volatile uint32_t *gpio, int gnn)
{
	int caa;
	uint32_t lev0, m;

	lev0 = *(gpio + GPLEV0);
	// printf("L0=0x%x\n", lev0); 

	caa = GPreadcap(gpio, gnn);

	// thegpinfo[j].char *pinno;
	// thegpinfo[j].char *fx[8]; 

	if(caa == 0 || caa == 1)
	{
		m = thegpinfo[gnn].ONEmask;

		printf("GPIO%02u pin %s %s (%s)\n", gnn, 
			thegpinfo[gnn].pinno, thegpinfo[gnn].fx[caa],
			(lev0 & m) ? "H" : "L");
	}
	else
		printf("GPIO%02u pin %s %s\n", gnn, 
			thegpinfo[gnn].pinno, thegpinfo[gnn].fx[caa]);

	return(gnn);
}

int GPshowfsel(volatile uint32_t *gpio)
{
	int j, gnn;

	for(j = 0; j<GPIOCOUNT; j++)
	{
		gnn = GPuseable[j];
		GPshowone(gpio, gnn);
	}

	return(0);
}

void GPset(volatile uint32_t *gpio, int number)
{
	*(gpio+GPSET0+thegpinfo[number].ONEoffset) = thegpinfo[number].ONEmask;
}

void GPclear(volatile uint32_t *gpio, int number)
{
	*(gpio+GPCLR0+thegpinfo[number].ONEoffset) = thegpinfo[number].ONEmask;
}

int GPread(volatile uint32_t *gpio, int number)
{
	int x;
	x = *(gpio+GPLEV0+thegpinfo[number].ONEoffset) & thegpinfo[number].ONEmask;
	return(x);
}



