

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdint.h>

/* For GPIO */
#include <sys/mman.h>
#include "kgpio.h"

/* For SPI */
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#include "mgraf.h"

/* Display connections:
	1: Vcc 3.3V
	2: GND
	3: SCE(L) - CS0(L) P1-24
	4: RST(L) - GPIO24 P1-18
	5: D/C -    GPIO25 P1-22
	6: MOSI   - MOSI   P1-19
	7: SCLK   - SCLK   P1-23
*/

#define MD_RESET 24
#define MD_D_C 25

/* 
	GPsetcap(volatile uint32_t *gpio, MD_RESET, GPIO_OUTPUT); 
	GPsetcap(volatile uint32_t *gpio, MD_D_C, GPIO_OUTPUT);

	Reset the device:
	GPset(volatile uint32_t *gpio, MD_RESET);
	GPclear(volatile uint32_t *gpio, MD_RESET);
	GPset(volatile uint32_t *gpio, MD_RESET);

	Select to write Commands:
	GPclear(volatile uint32_t *gpio, MD_D_C);

	Select to write Data:
	GPset(volatile uint32_t *gpio, MD_D_C);
*/

/* SPI:  Open, set mode, bits per word, speed.
	Display is write-only so we can use write() to push data to it.
	More advanced is the ioctl(SPI_IOC_MESSAGE(N)) for N messages, 
	each containing a transmission and a reception part.
*/


/* Size of LCD is 84 pixels wide and 48 pixels tall. */ 
#define LCD_X 84
#define LCD_Y 48

void displayexit(struct dispinfo *PDI)
{
	if(PDI->fd >= 0)
		close(PDI->fd);
}

int displayinit(struct dispinfo *PDI, char *infobuf, size_t ibmax)
{
	uint8_t mode;
	uint32_t speed;
	uint8_t bits;
	int oks;
	char spidevice[128];
	char errbuf[128];
	char cmd[8];

	PDI->fd = -1;
	PDI->gpio = NULL;

	oks = GPbasepointer(&(PDI->gpio));

	if(oks != 0)
	{
		strerror_r(errno, errbuf, 128);
		snprintf(infobuf, ibmax, "GPIO: %s", errbuf); 
		return(-1);
	}

	GPsetcap(PDI->gpio, MD_RESET, GPIO_OUTPUT); 
	GPsetcap(PDI->gpio, MD_D_C, GPIO_OUTPUT);

	strcpy(spidevice, "/dev/spidev0.0");

	PDI->fd = open(spidevice, O_RDWR);
	if(PDI->fd < 0)
	{
		strerror_r(errno, errbuf, 128);
		snprintf(infobuf, ibmax, "%s: %s", spidevice, errbuf); 
		return(-1);
	}

	/* The mode we will use here is mode 0:
		MSB first ?
		CS0 active Low
	*/
	mode = SPI_MODE_0; //  | SPI_LSB_FIRST;

	oks = ioctl(PDI->fd, SPI_IOC_WR_MODE, &mode);
	if(oks < 0)
	{
		strerror_r(errno, errbuf, 128);
		snprintf(infobuf, ibmax, "Setting device mode to 0x%x: %s", mode, errbuf); 
		close(PDI->fd);
		PDI->fd = -1;
		PDI->gpio = NULL;
		return(2);
	}

	bits = 8;
	oks = ioctl(PDI->fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
	if(oks < 0)
	{
		strerror_r(errno, errbuf, 128);
		snprintf(infobuf, ibmax, "Setting %d bits per word: %s", bits, errbuf); 
		close(PDI->fd);
		PDI->fd = -1;
		PDI->gpio = NULL;
		return(2);
	}

	speed = 100000;
	oks = ioctl(PDI->fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
	if(oks < 0)
	{
		strerror_r(errno, errbuf, 128);
		snprintf(infobuf, ibmax, "Setting speed to %u Hz: %s", speed, errbuf); 
		close(PDI->fd);
		PDI->fd = -1;
		PDI->gpio = NULL;
		return(2);
	}


	/* Reset */
	GPset(PDI->gpio, MD_RESET);
	GPclear(PDI->gpio, MD_RESET);
	GPset(PDI->gpio, MD_RESET);

	// GPset(PDI->gpio, MD_D_C); /* Write Data. */
	// GPclear(PDI->gpio, MD_D_C); /* Write Commands. */

	GPclear(PDI->gpio, MD_D_C); /* Write Commands. */
	cmd[0] = 0x21; /* Extended commands follow */
	cmd[1] = 0xB1; /* Set Vop, LCD Contrast (0x80-0xFF) 
						0xB1 may be good at 3.3V */
	cmd[2] = 0x04; /* Set Temp coefficient (4-7) */
	cmd[3] = 0x14; /* Set LCD bias mode, (0x10-0x17, 0x13 or 0x14) */

	cmd[4] = 0x20; /* Standard commands follow */

	cmd[5] = 0x0C; /* Display control mode: 00001D0E - 
						DE: 00 (0x08) is blank  
							01 (0x09) all bits on 
							10 (0x0C) normal
							11 (0x0D) inverse video. */ 

	write(PDI->fd, cmd, 6);

	/* Ready for action. Various commands are to be preceeded by 
		data or command selection. */

	return(0);
}


/* Start writing at address x,y indicated. */
void gotoxy(struct dispinfo *PDI, uint8_t x, uint8_t y)
{
	char buf[4];
	x &= 0x7F;	 /* 0 to 83 */
	y &= 0x7;	 /* 0 to 5 */

	GPclear(PDI->gpio, MD_D_C); /* Write Commands. */
	buf[0] = 0x80 | x;
	buf[1] = 0x40 | y;
	write(PDI->fd, buf, 2); 
}

/* Each byte maps to a cell 8 pixels tall so there are 6 rows
	with 8 cells in each, for 48 pixels vertical.

	Top pixel is the LSB. With normal video, bits 1 show pixels and bits
	0 do not. 

	So, let's see -- for meaningful graphics we will have to
	maintain a readable copy of the image in memory, and when drawing, 
	update those bytes that are affected.

	There is a total of 504 bytes, so the memory needed is modest.
	In addition we will want to keep some flags indicating need for
	update. (unless the gotoxy+write is more than just write 504,
	but that can be determined in each case: if too many bytes need
	updating scattered around, then just blast out the entire lot
	in one go.	504 bytes to write vs 3*updcount: 504/3 = 168 ... hmm).
*/

/* Draw/undraw a single pixel. */
/* Draw a v.line at x from y1 to y2. */
/* Draw a h.line from x1 to x2 at y. */
/* Draw strings of text at X,Y. */

/* Plot some data in a rectangular plot with X and Y */
/* What was the size of the font from CANIS and SYM7? */


/* Write all zeros. */
void clear(struct dispinfo *PDI)
{
	int j;
	char buf[4];

	gotoxy(PDI, 0, 0);

	GPset(PDI->gpio, MD_D_C); /* Write Data. */

	for(j = 0; j< (LCD_X * LCD_Y / 8); j++)
	{
		buf[0] = 0; 
		write(PDI->fd, buf, 1);
	}
}


/* Write all ones. */
void unclear(struct dispinfo *PDI)
{
	int j;
	char buf[4];

	GPset(PDI->gpio, MD_D_C); /* Write Data. */

	for(j = 0; j< (LCD_X * LCD_Y / 8); j++)
	{
		buf[0] = 0xFF; 
		write(PDI->fd, buf, 1);
	}
}

#if(0)

int main()
{
	struct dispinfo DI;
	int ds, j;
	char xpa[3];

	ds = displayinit(&DI);
	if(ds < 0)
	{
		fprintf(stderr, "%s\n", DI.infobuf);
		return(0);
	}
	fprintf(stderr, "About to clear...\n" );

	clear(&DI);
	sleep(1);

	gotoxy(&DI, 0, 0);
	GPset(DI.gpio, MD_D_C);     /* Write Data. */

	// xpa[0] = 0x55;
	// xpa[1] = 0xaa;

	for(j = 0; j<80; j++)
	{
		// write(DI.fd, xpa, 2);
		xpa[0] = j;
		write(DI.fd, xpa, 1);
	}

	// fprintf(stderr, "About to set...\n" );
	// unclear(&DI);
	// sleep(1);

	// fprintf(stderr, "About to clear again ...\n" );
	// clear(&DI);

	displayexit(&DI);
	return(0);
}


	// GPset(PDI->gpio, MD_D_C);     /* Write Data. */
	// GPclear(PDI->gpio, MD_D_C);     /* Write Command. */
	// extern uint8_t TXT5x8[][5] ;

#endif

