

/* kgpio.h -- look at GPIO lines.
	Needed along with SPI (SPIDEV) and i2c
	in combinations */


/* Uses the /dev/mem addresses to change GPIO lines. 
	Based to some extent on code seen in 

	http://elinux.org/RPi_Low-level_peripherals

	But there really ought to be some real device driver... 

	Further notes based on the datasheet published for the BCM2835.
*/

#ifndef KGPIOHINCLUDED
#define KGPIOHINCLUDED


#define BCM2708_PERI_BASE 0x20000000

/* According to the datasheet section 1.2.3 the MMU translates
	0x7Exx xxxx into the range 0x20xx xxxx, so we use this for
	the periferal base address. */

#define GPIO_BASE (BCM2708_PERI_BASE+0x200000)

/* Makes mapping to 0x7E20 0000 -- which is where GPIO registers are,
	as per datasheet section */

#define GPBLOCK_SIZE (4*1024)

/* GPIO function selections:

	Access bits in the 3-bit groups at +00 +04 +08 +0C +10 +14 
	(offsets 0 1 2 3 4 5, essentially the 10s digit of the bit number.) 
	Values for each of these triads:
	000 = input
	001 = output
	010 = Alternate function 5
	011 = Alternate function 4
	100 = Alternate function 0
	101 = Alternate function 1
	110 = Alternate function 2
	111 = Alternate function 3
*/
#define GPIO_INPUT 0
#define GPIO_OUTPUT 1

struct tag_gpioinfo
{
	int gpiono;			 /* GPIO number. */
	char *pinno;         /* Header and pin location. */
	char *fx[8];         /* the 8 possible functions. */
	int GPFSELindex;	 /* (offset for the GPFSEL). */
	uint32_t GPFSELmask; /* mask in GPFSEL. */
	int ONEoffset;		 /* Offset from the XXX0 register */
	uint32_t ONEmask;    /* mask for the 1-bit registers. */
};


/* Available gpios on the P1 and P5 headers: */ 
extern struct tag_gpioinfo thegpinfo[];

#define GPIOCOUNT 21

/* Useable on P1 and P5. */
extern int GPuseable[];


/* Output control: (Tables 6-8 onwards) */
#define GPSET0 7  /* +1C set outputs of gpio 0-31 */ 
#define GPSET1 8  /* +20 set outputs of gpio 32-53 (low 22 bits) */ 
#define GPCLR0 10 /* +28 clear outputs of gpio 0-31 */ 
#define GPCLR1 11 /* +2C clear outputs of gpio 32-53 (low 22 bits) */ 

// *(gpio+GPSET0+thegpinfo[gnn].ONEoffset) = thegpinfo[gnn].ONEmask;
// *(gpio+GPCLR0+thegpinfo[gnn].ONEoffset) = thegpinfo[gnn].ONEmask;

/* Input: (tables 6-12, 6-13)  */ 
#define GPLEV0 13 /* +34 read pin-level on gpio 0-31 */ 
#define GPLEV1 14 /* +38 read pin-level on gpio 32-53 (low 22 bits) */

/* Event status: (tables 6-14, 6-15)
	Bit values 1 mean there has been an event. 
	Write bit value 1 to clear the event flags. */
#define GPEDS0 16 /* +40 read event on gpio 0-31 */ 
#define GPEDS1 17 /* +44 read event on gpio 32-53 (low 22 bits) */ 

/* Define the event to be a sampled synchronized rising edge. 
	Set the bit to 1 for yes, 0 for no event on rising edge */ 
#define GPREN0 19 /* +4C gpio 0-31 */
#define GPREN1 20 /* +50 gpio 32-53 (low 22 bits) */

/* Define the event to be a sampled synchronized falling edge. 
	Set the bit to 1 for yes, 0 for no event on falling edge */
#define GPFEN0 22 /* +58 gpio 0-31 */
#define GPFEN1 23 /* +5C gpio 32-53 (low 22 bits) */

/* Define the event to be high level
	Set the bit to 1 for yes, 0 for no event on high level */ 
#define GPHEN0 25 /* +64 gpio 0-31 */
#define GPHEN1 26 /* +68 gpio 32-53 (low 22 bits) */

/* Define the event to be low level
	Set the bit to 1 for yes, 0 for no event on low level */
#define GPLEN0 28 /* +70 gpio 0-31 */
#define GPLEN1 29 /* +74 gpio 32-53 (low 22 bits) */

/* Define the event to be a asynchronous rising edge. 
	Set the bit to 1 for yes, 0 for no event on async rising edge */
#define GPAREN0 31 /* +7C gpio 0-31 */
#define GPAREN1 32 /* +80 gpio 32-53 (low 22 bits) */

/* Define the event to be an asynchronous falling edge. 
	Set the bit to 1 for yes, 0 for no event on async falling edge */
#define GPAFEN0 34 /* +88 gpio 0-31 */
#define GPAFEN1 35 /* +8C gpio 32-53 (low 22 bits) */


/* Pull-up or pull-down, applies to all IO lines. */
#define GPPUD 37 		/* +94 */
#define GPPUD_OFF 0
#define GPPUD_DOWN 1
#define GPPUD_UP 2

/* Pullup or -down clocking. 
	Bits set to 0 for no effect, 1 for assert clock on the line */

#define GPPUDCLK0 38 /* +98 gpio 0-31 */
#define GPPUDCLK1 39 /* +9C gpio 32-53 (low 22 bits) */

/* Find the base-pointer 
	Returns 0 for OK, -1 on failures.
*/

int GPbasepointer(volatile uint32_t **pgpio);

/* We have to do a munmap() of the gpio after use! 
	Bugs from not doing so have been seen, on the SDRs
	for example.
*/
int GPreleasebp(volatile uint32_t **pgpio);

/* Read the capabilty word for the gpio-line identified as channel. */
int GPreadcap(volatile uint32_t *gpio, int channel);

/* Set the function on the indicated channel to the value. Beware of
	illegal values ... most will only allow input or output. 
	Set to GPIO_INPUT or GPIO_OUTPUT as desired.  */

int GPsetcap(volatile uint32_t *gpio, int channel, int capability);

/* Show status of one GPIO line */
int GPshowone(volatile uint32_t *gpio, int gnn);

/* Show the functions and status on all the P1 and P5 GPIOs */
int GPshowfsel(volatile uint32_t *gpio);

/* Utility-functions for set, clear, and read. */
void GPset(volatile uint32_t *gpio, int number);

void GPclear(volatile uint32_t *gpio, int number);

int GPread(volatile uint32_t *gpio, int number);


#endif


